@@ -309,7 +309,7 @@ final class InjectorImpl implements Injector, Lookups {
   /** Returns true if the key type is MembersInjector (but not a subclass of MembersInjector). */
   private static boolean isMembersInjector(Key<?> key) {
     return key.getTypeLiteral().getRawType().equals(MembersInjector.class)
-        && !(key.getAnnotationType() != null);
+        && key.getAnnotationType() == null;
   }
 
   private <T> BindingImpl<MembersInjector<T>> createMembersInjectorBinding(
@@ -625,8 +625,10 @@ final class InjectorImpl implements Injector, Lookups {
       Errors errors, boolean jitBinding) throws ErrorsException {
     Class<?> rawType = key.getTypeLiteral().getRawType();
 
-    // Don't try to inject arrays, or enums.
-    if (rawType.isArray() || rawType.isEnum()) {
+    ImplementedBy implementedBy = rawType.getAnnotation(ImplementedBy.class);
+
+    // Don't try to inject arrays or enums annotated with @ImplementedBy.
+    if (rawType.isArray() || (rawType.isEnum() && implementedBy != null)) {
       throw errors.missingImplementation(key).toException();
     }
 
@@ -639,7 +641,6 @@ final class InjectorImpl implements Injector, Lookups {
     }
 
     // Handle @ImplementedBy
-    ImplementedBy implementedBy = rawType.getAnnotation(ImplementedBy.class);
     if (implementedBy != null) {
       Annotations.checkForMisplacedScopeAnnotations(rawType, source, errors);
       return createImplementedByBinding(key, scoping, implementedBy, errors);
@@ -1072,5 +1073,4 @@ final class InjectorImpl implements Injector, Lookups {
         .add("bindings", state.getExplicitBindingsThisLevel().values())
         .toString();
   }
-
 }
