@@ -16,6 +16,14 @@
 
 package com.google.inject.internal;
 
+import static com.google.inject.internal.WeakKeySetUtils.assertBlacklisted;
+import static com.google.inject.internal.WeakKeySetUtils.assertInSet;
+import static com.google.inject.internal.WeakKeySetUtils.assertNotBlacklisted;
+import static com.google.inject.internal.WeakKeySetUtils.assertNotInSet;
+import static com.google.inject.internal.WeakKeySetUtils.assertSourceNotInSet;
+import static com.google.inject.internal.WeakKeySetUtils.awaitClear;
+import static com.google.inject.internal.WeakKeySetUtils.awaitFullGc;
+
 import com.google.common.collect.ImmutableList;
 import com.google.common.collect.ImmutableMap;
 import com.google.common.collect.ImmutableSet;
@@ -42,6 +50,7 @@ import com.google.inject.spi.TypeListenerBinding;
 import junit.framework.TestCase;
 
 import java.lang.annotation.Annotation;
+import java.lang.ref.ReferenceQueue;
 import java.lang.ref.WeakReference;
 import java.util.Arrays;
 import java.util.List;
@@ -72,20 +81,17 @@ public class WeakKeySetTest extends TestCase {
     WeakReference<Key<Integer>> weakKeyRef = new WeakReference<Key<Integer>>(key);
 
     set.add(key, state, source);
-    assertTrue(set.contains(key));
-    assertEquals(1, set.getSources(key).size());
-    assertTrue(set.getSources(key).contains(source));
+    assertInSet(set, key, 1, source);
 
     state = null;
 
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
 
-    assertFalse(set.contains(Key.get(Integer.class)));
-    assertNull(set.getSources(Key.get(Integer.class)));
+    assertNotInSet(set, Key.get(Integer.class));
 
     // Ensure there are no hanging references.
     key = null;
-    GcFinalization.awaitClear(weakKeyRef);
+    awaitClear(weakKeyRef);
   }
   
   public void testEviction_nullSource() {
@@ -96,20 +102,17 @@ public class WeakKeySetTest extends TestCase {
     WeakReference<Key<Integer>> weakKeyRef = new WeakReference<Key<Integer>>(key);
 
     set.add(key, state, source);
-    assertTrue(set.contains(key));
-    assertEquals(1, set.getSources(key).size());
-    assertTrue(set.getSources(key).contains(source));
+    assertInSet(set, key, 1, source);
 
     state = null;
 
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
 
-    assertFalse(set.contains(Key.get(Integer.class)));
-    assertNull(set.getSources(Key.get(Integer.class)));
+    assertNotInSet(set, Key.get(Integer.class));
 
     // Ensure there are no hanging references.
     key = null;
-    GcFinalization.awaitClear(weakKeyRef);
+    awaitClear(weakKeyRef);
   }
 
   public void testEviction_keyOverlap_2x() {
@@ -121,14 +124,10 @@ public class WeakKeySetTest extends TestCase {
     Object source2 = new Object();
 
     set.add(key1, state1, source1);
-    assertTrue(set.contains(key1));
-    assertEquals(1, set.getSources(key1).size());
-    assertTrue(set.getSources(key1).contains(source1));
+    assertInSet(set, key1, 1, source1);
 
     set.add(key2, state2, source2);
-    assertTrue(set.contains(key2));
-    assertEquals(2, set.getSources(key2).size());
-    assertTrue(set.getSources(key2).containsAll(Arrays.asList(source1, source2)));
+    assertInSet(set, key2, 2, source1, source2);
 
     WeakReference<Key<Integer>> weakKey1Ref = new WeakReference<Key<Integer>>(key1);
     WeakReference<Key<Integer>> weakKey2Ref = new WeakReference<Key<Integer>>(key2);
@@ -138,16 +137,14 @@ public class WeakKeySetTest extends TestCase {
     Key<Integer> key = key1 = key2 = Key.get(Integer.class);
     state1 = null;
 
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
 
-    assertTrue(set.contains(key));
-    assertEquals(1, set.getSources(key).size());
-    assertTrue(set.getSources(key).contains(source2));
-    assertFalse(set.getSources(key).contains(source1));
+    assertSourceNotInSet(set, key, source1);
+    assertInSet(set, key, 1, source2);
 
     source1 = source2 = null;
     
-    GcFinalization.awaitClear(weakSource1Ref);
+    awaitClear(weakSource1Ref);
     // Key1 will be referenced as the key in the sources backingSet and won't be
     // GC'd.
     
@@ -156,15 +153,14 @@ public class WeakKeySetTest extends TestCase {
 
     state2 = null;
 
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
 
-    assertFalse(set.contains(key));
-    assertNull(set.getSources(key));
+    assertNotInSet(set, key);
 
-    GcFinalization.awaitClear(weakKey2Ref);
-    GcFinalization.awaitClear(weakSource2Ref);
+    awaitClear(weakKey2Ref);
+    awaitClear(weakSource2Ref);
     // Now that the backing set is emptied, key1 is released.
-    GcFinalization.awaitClear(weakKey1Ref);
+    awaitClear(weakKey1Ref);
   }
   
   public void testNoEviction_keyOverlap_2x() {
@@ -176,25 +172,18 @@ public class WeakKeySetTest extends TestCase {
     Object source2 = new Object();
 
     set.add(key1, state1, source1);
-    assertTrue(set.contains(key1));
-    assertEquals(1, set.getSources(key1).size());
-    assertTrue(set.getSources(key1).contains(source1));
+    assertInSet(set, key1, 1, source1);
 
     set.add(key2, state2, source2);
-    assertTrue(set.contains(key2));
-    assertEquals(2, set.getSources(key2).size());
-    assertTrue(set.getSources(key1).containsAll(Arrays.asList(source1, source2)));
+    assertInSet(set, key2, 2, source1, source2);
 
     WeakReference<Key<Integer>> weakKey1Ref = new WeakReference<Key<Integer>>(key1);
     WeakReference<Key<Integer>> weakKey2Ref = new WeakReference<Key<Integer>>(key2);
 
     Key<Integer> key = key1 = key2 = Key.get(Integer.class);
 
-    GcFinalization.awaitFullGc();
-
-    assertTrue(set.contains(key));
-    assertEquals(2, set.getSources(key).size());
-    assertTrue(set.getSources(key1).containsAll(Arrays.asList(source1, source2)));
+    awaitFullGc();
+    assertInSet(set, key, 2, source1, source2);
 
     // Ensure the keys don't get GC'd when states are still referenced. key1 will be present in the
     // as the map key but key2 could be GC'd if the implementation does something wrong.
@@ -210,16 +199,12 @@ public class WeakKeySetTest extends TestCase {
     Object source = null;
 
     set.add(key1, state1, source);
-    assertTrue(set.contains(key1));
-    assertEquals(1, set.getSources(key1).size());
-    assertTrue(set.getSources(key1).contains(source));
+    assertInSet(set, key1, 1, source);
 
     set.add(key2, state2, source);
-    assertTrue(set.contains(key2));
-
     // Same source so still only one value.
-    assertEquals(1, set.getSources(key2).size());
-    assertTrue(set.getSources(key1).contains(source));
+    assertInSet(set, key2, 1, source);
+    assertInSet(set, key1, 1, source);
 
     WeakReference<Key<Integer>> weakKey1Ref = new WeakReference<Key<Integer>>(key1);
     WeakReference<Key<Integer>> weakKey2Ref = new WeakReference<Key<Integer>>(key2);
@@ -228,31 +213,25 @@ public class WeakKeySetTest extends TestCase {
     Key<Integer> key = key1 = key2 = Key.get(Integer.class);
     state1 = null;
 
-    GcFinalization.awaitFullGc();
-
-    assertTrue(set.contains(key));
-
+    awaitFullGc();
     // Should still have a single source.
-    assertEquals(1, set.getSources(key).size());
-    assertTrue(set.getSources(key1).contains(source));
-    
+    assertInSet(set, key, 1, source);
+
     source = null;
 
-    GcFinalization.awaitClear(weakSourceRef);
+    awaitClear(weakSourceRef);
     // Key1 will be referenced as the key in the sources backingSet and won't be
     // GC'd.
 
     state2 = null;
 
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
+    assertNotInSet(set, key);
 
-    assertFalse(set.contains(key));
-    assertNull(set.getSources(key));
-
-    GcFinalization.awaitClear(weakKey2Ref);
-    GcFinalization.awaitClear(weakSourceRef);
+    awaitClear(weakKey2Ref);
+    awaitClear(weakSourceRef);
     // Now that the backing set is emptied, key1 is released.
-    GcFinalization.awaitClear(weakKey1Ref);
+    awaitClear(weakKey1Ref);
   }
   
   public void testEviction_keyAndSourceOverlap_nonNull() {
@@ -263,16 +242,11 @@ public class WeakKeySetTest extends TestCase {
     Object source = new Object();
 
     set.add(key1, state1, source);
-    assertTrue(set.contains(key1));
-    assertEquals(1, set.getSources(key1).size());
-    assertTrue(set.getSources(key1).contains(source));
+    assertInSet(set, key1, 1, source);
 
     set.add(key2, state2, source);
-    assertTrue(set.contains(key2));
-
     // Same source so still only one value.
-    assertEquals(1, set.getSources(key2).size());
-    assertTrue(set.getSources(key1).contains(source));
+    assertInSet(set, key2, 1, source);
 
     WeakReference<Key<Integer>> weakKey1Ref = new WeakReference<Key<Integer>>(key1);
     WeakReference<Key<Integer>> weakKey2Ref = new WeakReference<Key<Integer>>(key2);
@@ -281,32 +255,29 @@ public class WeakKeySetTest extends TestCase {
     Key<Integer> key = key1 = key2 = Key.get(Integer.class);
     state1 = null;
 
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
 
-    assertTrue(set.contains(key));
-
-    // Should still have a single source.
-    assertEquals(1, set.getSources(key).size());
-    assertTrue(set.getSources(key1).contains(source));
+ // Same source so still only one value.
+    assertInSet(set, key, 1, source);
+    assertInSet(set, key1, 1, source);
     
     source = null;
 
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
     assertNotNull(weakSourceRef.get());
     // Key1 will be referenced as the key in the sources backingSet and won't be
     // GC'd.
 
     state2 = null;
 
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
 
-    assertFalse(set.contains(key));
-    assertNull(set.getSources(key));
+    assertNotInSet(set, key);
 
-    GcFinalization.awaitClear(weakKey2Ref);
-    GcFinalization.awaitClear(weakSourceRef);
+    awaitClear(weakKey2Ref);
+    awaitClear(weakSourceRef);
     // Now that the backing set is emptied, key1 is released.
-    GcFinalization.awaitClear(weakKey1Ref);
+    awaitClear(weakKey1Ref);
   }
 
   public void testEviction_keyOverlap_3x() {
@@ -321,19 +292,13 @@ public class WeakKeySetTest extends TestCase {
     Object source3 = new Object();
 
     set.add(key1, state1, source1);
-    assertTrue(set.contains(key1));
-    assertEquals(1, set.getSources(key1).size());
-    assertTrue(set.getSources(key1).contains(source1));
+    assertInSet(set, key1, 1, source1);
 
     set.add(key2, state2, source2);
-    assertTrue(set.contains(key2));
-    assertEquals(2, set.getSources(key2).size());
-    assertTrue(set.getSources(key1).containsAll(Arrays.asList(source1, source2)));
+    assertInSet(set, key1, 2, source1, source2);
 
     set.add(key3, state3, source3);
-    assertTrue(set.contains(key3));
-    assertEquals(3, set.getSources(key3).size());
-    assertTrue(set.getSources(key1).containsAll(Arrays.asList(source1, source2, source3)));
+    assertInSet(set, key1, 3, source1, source2, source3);
 
     WeakReference<Key<Integer>> weakKey1Ref = new WeakReference<Key<Integer>>(key1);
     WeakReference<Key<Integer>> weakKey2Ref = new WeakReference<Key<Integer>>(key2);
@@ -345,42 +310,36 @@ public class WeakKeySetTest extends TestCase {
     Key<Integer> key = key1 = key2 = key3 = Key.get(Integer.class);
     state1 = null;
 
-    GcFinalization.awaitFullGc();
-
-    assertTrue(set.contains(key));
-    assertEquals(2, set.getSources(key).size());
-    assertTrue(set.getSources(key).containsAll(Arrays.asList(source2, source3)));
+    awaitFullGc();
+    assertSourceNotInSet(set, key, source1);
+    assertInSet(set, key, 2, source2, source3);
 
     source1 = null;
     // Key1 will be referenced as the key in the sources backingSet and won't be
     // GC'd.
-    GcFinalization.awaitClear(weakSource1Ref);
+    awaitClear(weakSource1Ref);
 
     state2 = null;
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
+    assertSourceNotInSet(set, key, source2);
+    assertInSet(set, key, 1, source3);
 
-    assertTrue(set.contains(key));
-    assertEquals(1, set.getSources(key).size());
-    assertTrue(set.getSources(key).contains(source3));
-
-    GcFinalization.awaitClear(weakKey2Ref);
+    awaitClear(weakKey2Ref);
     
     source2 = null;
-    GcFinalization.awaitClear(weakSource2Ref);
+    awaitClear(weakSource2Ref);
     // Key1 will be referenced as the key in the sources backingSet and won't be
     // GC'd.
 
     state3 = null;
-    GcFinalization.awaitFullGc();
+    awaitFullGc();
+    assertNotInSet(set, key);
 
-    assertFalse(set.contains(Key.get(Integer.class)));
-    assertNull(set.getSources(Key.get(Integer.class)));
-
-    GcFinalization.awaitClear(weakKey3Ref);
+    awaitClear(weakKey3Ref);
     source3 = null;
-    GcFinalization.awaitClear(weakSource3Ref);
+    awaitClear(weakSource3Ref);
     // Now that the backing set is emptied, key1 is released.
-    GcFinalization.awaitClear(weakKey1Ref);
+    awaitClear(weakKey1Ref);
   }
 
   public void testWeakKeySet_integration() {
@@ -401,7 +360,7 @@ public class WeakKeySetTest extends TestCase {
     
     // Clear the ref, GC, and ensure that we are no longer blacklisting.
     childInjector = null;
-    GcFinalization.awaitClear(weakRef);
+    awaitClear(weakRef);
     assertNotBlacklisted(parentInjector, Key.get(String.class));
   }
   
@@ -434,13 +393,13 @@ public class WeakKeySetTest extends TestCase {
     
     // Clear ref1, GC, and ensure that we still blacklist.
     childInjector1 = null;
-    GcFinalization.awaitClear(weakRef1);
+    awaitClear(weakRef1);
     assertNotBlacklisted(parentInjector, Key.get(String.class));
     assertBlacklisted(parentInjector, Key.get(Long.class));
 
     // Clear the ref, GC, and ensure that we are no longer blacklisting.
     childInjector2 = null;
-    GcFinalization.awaitClear(weakRef2);
+    awaitClear(weakRef2);
     assertNotBlacklisted(parentInjector, Key.get(String.class));
     assertNotBlacklisted(parentInjector, Key.get(Long.class));
   }
@@ -471,27 +430,15 @@ public class WeakKeySetTest extends TestCase {
     
     // Clear ref1, GC, and ensure that we still blacklist.
     childInjector1 = null;
-    GcFinalization.awaitClear(weakRef1);
+    awaitClear(weakRef1);
     assertBlacklisted(parentInjector, Key.get(String.class));
 
     // Clear the ref, GC, and ensure that we are no longer blacklisting.
     childInjector2 = null;
-    GcFinalization.awaitClear(weakRef2);
+    awaitClear(weakRef2);
     assertNotBlacklisted(parentInjector, Key.get(String.class));
   }
-  
-  private static void assertBlacklisted(Injector injector, Key<?> key) {
-    assertBlacklistState(injector, key, true);
-  }
-  
-  private static void assertNotBlacklisted(Injector injector, Key<?> key) {
-    assertBlacklistState(injector, key, false);
-  }
 
-  private static void assertBlacklistState(Injector injector, Key<?> key, boolean isBlacklisted) {
-    assertEquals(isBlacklisted, ((InjectorImpl) injector).state.isBlacklisted(key));
-  }
-  
   private static class TestState implements State {
     public State parent() {
       return new TestState();
