@@ -16,8 +16,6 @@
 
 package com.google.inject.multibindings;
 
-import static com.google.inject.internal.RehashableKeys.Keys.needsRehashing;
-import static com.google.inject.internal.RehashableKeys.Keys.rehash;
 import static com.google.inject.multibindings.Element.Type.MAPBINDER;
 import static com.google.inject.multibindings.Multibinder.checkConfiguration;
 import static com.google.inject.multibindings.Multibinder.checkNotNull;
@@ -25,8 +23,10 @@ import static com.google.inject.multibindings.Multibinder.setOf;
 import static com.google.inject.util.Types.newParameterizedType;
 import static com.google.inject.util.Types.newParameterizedTypeWithOwner;
 
-import com.google.common.base.Supplier;
 import com.google.common.base.Joiner;
+import com.google.common.base.Objects;
+import com.google.common.base.Supplier;
+import com.google.common.collect.HashMultimap;
 import com.google.common.collect.ImmutableList;
 import com.google.common.collect.ImmutableMap;
 import com.google.common.collect.ImmutableSet;
@@ -45,9 +45,11 @@ import com.google.inject.Provider;
 import com.google.inject.TypeLiteral;
 import com.google.inject.binder.LinkedBindingBuilder;
 import com.google.inject.internal.Errors;
+import com.google.inject.multibindings.Indexer.IndexedBinding;
 import com.google.inject.multibindings.Multibinder.RealMultibinder;
 import com.google.inject.spi.BindingTargetVisitor;
 import com.google.inject.spi.Dependency;
+import com.google.inject.spi.Element;
 import com.google.inject.spi.HasDependencies;
 import com.google.inject.spi.ProviderInstanceBinding;
 import com.google.inject.spi.ProviderLookup;
@@ -364,92 +366,26 @@ public abstract class MapBinder<K, V> {
       checkNotNull(key, "key");
       checkConfiguration(!isInitialized(), "MapBinder was already initialized");
 
-      RealElement.BindingBuilder<V> valueBinding = RealElement.addMapBinding(
-          binder, key, keyType, valueType, entrySetBinder.getSetName());
-      Key<V> valueKey = Key.get(valueType, valueBinding.getAnnotation());
+      Key<V> valueKey = Key.get(valueType,
+          new RealElement(entrySetBinder.getSetName(), MAPBINDER, keyType.toString()));
       entrySetBinder.addBinding().toProvider(new ProviderMapEntry<K, V>(
           key, binder.getProvider(valueKey), valueKey));
-      return valueBinding;
+      return binder.bind(valueKey);
     }
 
-    public void configure(Binder binder) {
+    @Override public void configure(Binder binder) {
       checkConfiguration(!isInitialized(), "MapBinder was already initialized");
 
-      final ImmutableSet<Dependency<?>> dependencies
+      ImmutableSet<Dependency<?>> dependencies
           = ImmutableSet.<Dependency<?>>of(Dependency.get(entrySetBinder.getSetKey()));
 
       // Binds a Map<K, Provider<V>> from a collection of Set<Entry<K, Provider<V>>.
-      final Provider<Set<Entry<K, Provider<V>>>> entrySetProvider = binder
+      Provider<Set<Entry<K, Provider<V>>>> entrySetProvider = binder
           .getProvider(entrySetBinder.getSetKey());
-      
+
       binder.bind(providerMapKey).toProvider(
-          new RealMapBinderProviderWithDependencies<Map<K, Provider<V>>>(mapKey) {
-        private Map<K, Provider<V>> providerMap;
+          new RealProviderMapProvider(dependencies, entrySetProvider));
 
-        @Toolable @Inject void initialize(Injector injector) {
-          RealMapBinder.this.binder = null;
-          permitDuplicates = entrySetBinder.permitsDuplicates(injector);
-
-          Map<K, Provider<V>> providerMapMutable = new LinkedHashMap<K, Provider<V>>();
-          List<Map.Entry<K, Binding<V>>> bindingsMutable = Lists.newArrayList();
-          Set<K> duplicateKeys = null;
-          for (Entry<K, Provider<V>> entry : entrySetProvider.get()) {
-            Provider<V> previous = providerMapMutable.put(entry.getKey(), entry.getValue());
-            if (previous != null && !permitDuplicates) {
-              if (duplicateKeys == null) {
-                duplicateKeys = Sets.newHashSet();
-              }
-              duplicateKeys.add(entry.getKey());
-            }
-            ProviderMapEntry<K, V> providerEntry = (ProviderMapEntry<K, V>) entry;
-            Key<V> valueKey = providerEntry.getValueKey();
-            bindingsMutable.add(Maps.immutableEntry(entry.getKey(), injector.getBinding(valueKey)));
-          }
-          if (duplicateKeys != null) {
-            // Must use a ListMultimap in case more than one binding has the same source
-            // and is listed multiple times.
-            Multimap<K, String> dups = newLinkedKeyArrayValueMultimap();
-            for (Map.Entry<K, Binding<V>> entry : bindingsMutable) {
-              if (duplicateKeys.contains(entry.getKey())) {
-                dups.put(entry.getKey(), "\t at " + Errors.convert(entry.getValue().getSource()));
-              }
-            }
-            StringBuilder sb = new StringBuilder("Map injection failed due to duplicated key ");
-            boolean first = true;
-            for (K key : dups.keySet()) {
-              if (first) {
-                first = false;
-                if (duplicateKeyErrorMessages.containsKey(key)) {
-                  sb.setLength(0);
-                  sb.append(duplicateKeyErrorMessages.get(key));
-                } else {
-                  sb.append("\"" + key + "\", from bindings:\n");
-                }
-              } else {
-                if (duplicateKeyErrorMessages.containsKey(key)) {
-                  sb.append("\n and " + duplicateKeyErrorMessages.get(key));
-                } else {
-                  sb.append("\n and key: \"" + key + "\", from bindings:\n");
-                }
-              }
-              Joiner.on('\n').appendTo(sb, dups.get(key)).append("\n");
-            }
-            checkConfiguration(false, sb.toString());
-          }
-
-          providerMap = ImmutableMap.copyOf(providerMapMutable);
-          mapBindings = ImmutableList.copyOf(bindingsMutable);
-        }
-
-        public Map<K, Provider<V>> get() {
-          return providerMap;
-        }
-
-        public Set<Dependency<?>> getDependencies() {
-          return dependencies;
-        }
-      });
-      
       // The map this exposes is internally an ImmutableMap, so it's OK to massage
       // the guice Provider to javax Provider in the value (since Guice provider
       // implements javax Provider).
@@ -457,70 +393,11 @@ public abstract class MapBinder<K, V> {
       Key massagedProviderMapKey = (Key)providerMapKey;
       binder.bind(javaxProviderMapKey).to(massagedProviderMapKey);
 
-      final Provider<Map<K, Provider<V>>> mapProvider = binder.getProvider(providerMapKey);
-      binder.bind(mapKey).toProvider(new RealMapWithExtensionProvider<Map<K, V>>(mapKey) {
-        public Map<K, V> get() {
-          Map<K, V> map = new LinkedHashMap<K, V>();
-          for (Entry<K, Provider<V>> entry : mapProvider.get().entrySet()) {
-            V value = entry.getValue().get();
-            K key = entry.getKey();
-            checkConfiguration(value != null,
-                "Map injection failed due to null value for key \"%s\"", key);
-            map.put(key, value);
-          }
-          return Collections.unmodifiableMap(map);
-        }
-
-        public Set<Dependency<?>> getDependencies() {
-          return dependencies;
-        }
-
-        @SuppressWarnings("unchecked")
-        public <B, R> R acceptExtensionVisitor(BindingTargetVisitor<B, R> visitor,
-            ProviderInstanceBinding<? extends B> binding) {
-          if (visitor instanceof MultibindingsTargetVisitor) {
-            return ((MultibindingsTargetVisitor<Map<K, V>, R>)visitor).visit(this);
-          } else {
-            return visitor.visit(binding);
-          }
-        }
-
-        public Key<Map<K, V>> getMapKey() {
-          return mapKey;
-        }
-
-        public TypeLiteral<?> getKeyTypeLiteral() {
-          return keyType;
-        }
-
-        public TypeLiteral<?> getValueTypeLiteral() {
-          return valueType;
-        }
-
-        @SuppressWarnings("unchecked")
-        public List<Entry<?, Binding<?>>> getEntries() {
-          if (isInitialized()) {
-            return (List)mapBindings; // safe because mapBindings is immutable
-          } else {
-            throw new UnsupportedOperationException("getElements() not supported for module bindings");
-          }
-        }
-
-        public boolean permitsDuplicates() {
-          if (isInitialized()) {
-            return permitDuplicates;
-          } else {
-            throw new UnsupportedOperationException("permitsDuplicates() not supported for module bindings");
-          }
-        }
-
-        public boolean containsElement(com.google.inject.spi.Element element) {
-          return RealMapBinder.this.containsElement(element);
-        }
-      });
+      Provider<Map<K, Provider<V>>> mapProvider = binder.getProvider(providerMapKey);
+      binder.bind(mapKey).toProvider(new RealMapProvider(dependencies, mapProvider));
     }
 
-    boolean containsElement(com.google.inject.spi.Element element) {
+    boolean containsElement(Element element) {
       if (entrySetBinder.containsElement(element)) {
         return true;
       } else {
@@ -548,7 +425,7 @@ public abstract class MapBinder<K, V> {
       return key.getAnnotation() instanceof RealElement
           && ((RealElement) key.getAnnotation()).setName().equals(entrySetBinder.getSetName())
           && ((RealElement) key.getAnnotation()).type() == MAPBINDER
-          && ((RealElement) key.getAnnotation()).mapKeyType.equals(keyType)
+          && ((RealElement) key.getAnnotation()).keyType().equals(keyType.toString())
           && key.getTypeLiteral().equals(valueType);
     }
 
@@ -565,10 +442,174 @@ public abstract class MapBinder<K, V> {
       return mapKey.hashCode();
     }
 
+    final class RealProviderMapProvider
+        extends RealMapBinderProviderWithDependencies<Map<K, Provider<V>>> {
+      private final ImmutableSet<Dependency<?>> dependencies;
+      private final Provider<Set<Entry<K, Provider<V>>>> entrySetProvider;
+      private Map<K, Provider<V>> providerMap;
+
+      private RealProviderMapProvider(
+          ImmutableSet<Dependency<?>> dependencies,
+          Provider<Set<Entry<K, Provider<V>>>> entrySetProvider) {
+        super(mapKey);
+        this.dependencies = dependencies;
+        this.entrySetProvider = entrySetProvider;
+      }
+
+      @Toolable @Inject void initialize(Injector injector) {
+        RealMapBinder.this.binder = null;
+        permitDuplicates = entrySetBinder.permitsDuplicates(injector);
+
+        Map<K, Provider<V>> providerMapMutable = new LinkedHashMap<K, Provider<V>>();
+        List<Map.Entry<K, Binding<V>>> bindingsMutable = Lists.newArrayList();
+        Indexer indexer = new Indexer(injector);
+        Multimap<K, IndexedBinding> index = HashMultimap.create();
+        Set<K> duplicateKeys = null;
+        for (Entry<K, Provider<V>> entry : entrySetProvider.get()) {
+          ProviderMapEntry<K, V> providerEntry = (ProviderMapEntry<K, V>) entry;
+          Key<V> valueKey = providerEntry.getValueKey();
+          Binding<V> valueBinding = injector.getBinding(valueKey);
+          // If this isn't a dup due to an exact same binding, add it.
+          if (index.put(entry.getKey(), valueBinding.acceptTargetVisitor(indexer))) {
+            Provider<V> previous = providerMapMutable.put(entry.getKey(), entry.getValue());
+            if (previous != null && !permitDuplicates) {
+              if (duplicateKeys == null) {
+                duplicateKeys = Sets.newHashSet();
+              }
+              duplicateKeys.add(entry.getKey());
+            }
+            bindingsMutable.add(Maps.immutableEntry(entry.getKey(), valueBinding));
+          }
+        }
+        if (duplicateKeys != null) {
+          // Must use a ListMultimap in case more than one binding has the same source
+          // and is listed multiple times.
+          Multimap<K, String> dups = newLinkedKeyArrayValueMultimap();
+          for (Map.Entry<K, Binding<V>> entry : bindingsMutable) {
+            if (duplicateKeys.contains(entry.getKey())) {
+              dups.put(entry.getKey(), "\t at " + Errors.convert(entry.getValue().getSource()));
+            }
+          }
+          StringBuilder sb = new StringBuilder("Map injection failed due to duplicated key ");
+          boolean first = true;
+          for (K key : dups.keySet()) {
+            if (first) {
+              first = false;
+              if (duplicateKeyErrorMessages.containsKey(key)) {
+                sb.setLength(0);
+                sb.append(duplicateKeyErrorMessages.get(key));
+              } else {
+                sb.append("\"" + key + "\", from bindings:\n");
+              }
+            } else {
+              if (duplicateKeyErrorMessages.containsKey(key)) {
+                sb.append("\n and " + duplicateKeyErrorMessages.get(key));
+              } else {
+                sb.append("\n and key: \"" + key + "\", from bindings:\n");
+              }
+            }
+            Joiner.on('\n').appendTo(sb, dups.get(key)).append("\n");
+          }
+          checkConfiguration(false, sb.toString());
+        }
+
+        providerMap = ImmutableMap.copyOf(providerMapMutable);
+        mapBindings = ImmutableList.copyOf(bindingsMutable);
+      }
+
+      @Override public Map<K, Provider<V>> get() {
+        return providerMap;
+      }
+
+      @Override public Set<Dependency<?>> getDependencies() {
+        return dependencies;
+      }
+    }
+
+    final class RealMapProvider extends RealMapWithExtensionProvider<Map<K, V>> {
+      private final ImmutableSet<Dependency<?>> dependencies;
+      private final Provider<Map<K, Provider<V>>> mapProvider;
+
+      private RealMapProvider(
+          ImmutableSet<Dependency<?>> dependencies,
+          Provider<Map<K, Provider<V>>> mapProvider) {
+        super(mapKey);
+        this.dependencies = dependencies;
+        this.mapProvider = mapProvider;
+      }
+
+      @Override public Map<K, V> get() {
+        // We can initialize the internal table efficiently this way and then swap the values
+        // one by one.
+        Map<K, Object> map = new LinkedHashMap<K, Object>(mapProvider.get());
+        for (Entry<K, Object> entry : map.entrySet()) {
+          @SuppressWarnings("unchecked")  // we initialized the entries with providers
+          V value = ((Provider<V>) entry.getValue()).get();
+          checkConfiguration(value != null,
+              "Map injection failed due to null value for key \"%s\"", entry.getKey());
+          entry.setValue(value);
+        }
+        @SuppressWarnings("unchecked")  // if we exited the loop then we replaced all Providers
+        Map<K, V> typedMap = (Map<K, V>) map;
+        return Collections.unmodifiableMap(typedMap);
+      }
+
+      @Override public Set<Dependency<?>> getDependencies() {
+        return dependencies;
+      }
+
+      @SuppressWarnings("unchecked")
+      @Override
+      public <B, R> R acceptExtensionVisitor(BindingTargetVisitor<B, R> visitor,
+          ProviderInstanceBinding<? extends B> binding) {
+        if (visitor instanceof MultibindingsTargetVisitor) {
+          return ((MultibindingsTargetVisitor<Map<K, V>, R>)visitor).visit(this);
+        } else {
+          return visitor.visit(binding);
+        }
+      }
+
+      @Override public Key<Map<K, V>> getMapKey() {
+        return mapKey;
+      }
+
+      @Override public TypeLiteral<?> getKeyTypeLiteral() {
+        return keyType;
+      }
+
+      @Override public TypeLiteral<?> getValueTypeLiteral() {
+        return valueType;
+      }
+
+      @SuppressWarnings("unchecked")
+      @Override
+      public List<Entry<?, Binding<?>>> getEntries() {
+        if (isInitialized()) {
+          return (List)mapBindings; // safe because mapBindings is immutable
+        } else {
+          throw new UnsupportedOperationException(
+              "getElements() not supported for module bindings");
+        }
+      }
+
+      @Override public boolean permitsDuplicates() {
+        if (isInitialized()) {
+          return permitDuplicates;
+        } else {
+          throw new UnsupportedOperationException(
+              "permitsDuplicates() not supported for module bindings");
+        }
+      }
+
+      @Override public boolean containsElement(Element element) {
+        return RealMapBinder.this.containsElement(element);
+      }
+    }
+
     /**
      * Binds {@code Map<K, Set<V>>} and {{@code Map<K, Set<Provider<V>>>}.
      */
-    private static final class MultimapBinder<K, V> implements Module {
+    static final class MultimapBinder<K, V> implements Module {
 
       private final Key<Map<K, Set<V>>> multimapKey;
       private final Key<Map<K, Set<Provider<V>>>> providerMultimapKey;
@@ -583,72 +624,21 @@ public abstract class MapBinder<K, V> {
         this.entrySetKey = entrySetKey;
       }
 
-      public void configure(Binder binder) {
-        final ImmutableSet<Dependency<?>> dependencies
+      @Override public void configure(Binder binder) {
+        ImmutableSet<Dependency<?>> dependencies
             = ImmutableSet.<Dependency<?>>of(Dependency.get(entrySetKey));
 
-        final Provider<Set<Entry<K, Provider<V>>>> entrySetProvider =
+        Provider<Set<Entry<K, Provider<V>>>> entrySetProvider =
             binder.getProvider(entrySetKey);
         // Binds a Map<K, Set<Provider<V>>> from a collection of Map<Entry<K, Provider<V>> if
         // permitDuplicates was called.
         binder.bind(providerMultimapKey).toProvider(
-            new RealMapBinderProviderWithDependencies<Map<K, Set<Provider<V>>>>(multimapKey) {
-              private Map<K, Set<Provider<V>>> providerMultimap;
+            new RealProviderMultimapProvider(dependencies, entrySetProvider));
 
-              @SuppressWarnings("unused")
-              @Inject void initialize(Injector injector) {
-                Map<K, ImmutableSet.Builder<Provider<V>>> providerMultimapMutable =
-                    new LinkedHashMap<K, ImmutableSet.Builder<Provider<V>>>();
-                for (Entry<K, Provider<V>> entry : entrySetProvider.get()) {
-                  if (!providerMultimapMutable.containsKey(entry.getKey())) {
-                    providerMultimapMutable.put(
-                        entry.getKey(), ImmutableSet.<Provider<V>>builder());
-                  }
-                  providerMultimapMutable.get(entry.getKey()).add(entry.getValue());
-                }
-
-                ImmutableMap.Builder<K, Set<Provider<V>>> providerMultimapBuilder =
-                    ImmutableMap.builder();
-                for (Entry<K, ImmutableSet.Builder<Provider<V>>> entry
-                    : providerMultimapMutable.entrySet()) {
-                  providerMultimapBuilder.put(entry.getKey(), entry.getValue().build());
-                }
-                providerMultimap = providerMultimapBuilder.build();
-              }
-
-              public Map<K, Set<Provider<V>>> get() {
-                return providerMultimap;
-              }
-
-              public Set<Dependency<?>> getDependencies() {
-                return dependencies;
-              }
-            });
-
-        final Provider<Map<K, Set<Provider<V>>>> multimapProvider =
+        Provider<Map<K, Set<Provider<V>>>> multimapProvider =
             binder.getProvider(providerMultimapKey);
-        binder.bind(multimapKey).toProvider(new RealMapBinderProviderWithDependencies<Map<K, Set<V>>>(multimapKey) {
-
-          public Map<K, Set<V>> get() {
-            ImmutableMap.Builder<K, Set<V>> multimapBuilder = ImmutableMap.builder();
-            for (Entry<K, Set<Provider<V>>> entry : multimapProvider.get().entrySet()) {
-              K key = entry.getKey();
-              ImmutableSet.Builder<V> valuesBuilder = ImmutableSet.builder();
-              for (Provider<V> valueProvider : entry.getValue()) {
-                V value = valueProvider.get();
-                checkConfiguration(value != null,
-                    "Multimap injection failed due to null value for key \"%s\"", key);
-                valuesBuilder.add(value);
-              }
-              multimapBuilder.put(key, valuesBuilder.build());
-            }
-            return multimapBuilder.build();
-          }
-
-          public Set<Dependency<?>> getDependencies() {
-            return dependencies;
-          }
-        });
+        binder.bind(multimapKey).toProvider(
+            new RealMultimapProvider(dependencies, multimapProvider));
       }
 
       @Override public int hashCode() {
@@ -659,80 +649,141 @@ public abstract class MapBinder<K, V> {
         return o instanceof MultimapBinder
             && ((MultimapBinder<?, ?>) o).multimapKey.equals(multimapKey);
       }
+
+      final class RealProviderMultimapProvider
+          extends RealMapBinderProviderWithDependencies<Map<K, Set<Provider<V>>>> {
+        private final ImmutableSet<Dependency<?>> dependencies;
+        private final Provider<Set<Entry<K, Provider<V>>>> entrySetProvider;
+        private Map<K, Set<Provider<V>>> providerMultimap;
+
+        private RealProviderMultimapProvider(ImmutableSet<Dependency<?>> dependencies,
+            Provider<Set<Entry<K, Provider<V>>>> entrySetProvider) {
+          super(multimapKey);
+          this.dependencies = dependencies;
+          this.entrySetProvider = entrySetProvider;
+        }
+
+        @SuppressWarnings("unused")
+        @Inject void initialize(Injector injector) {
+          Map<K, ImmutableSet.Builder<Provider<V>>> providerMultimapMutable =
+              new LinkedHashMap<K, ImmutableSet.Builder<Provider<V>>>();
+          for (Entry<K, Provider<V>> entry : entrySetProvider.get()) {
+            if (!providerMultimapMutable.containsKey(entry.getKey())) {
+              providerMultimapMutable.put(
+                  entry.getKey(), ImmutableSet.<Provider<V>>builder());
+            }
+            providerMultimapMutable.get(entry.getKey()).add(entry.getValue());
+          }
+
+          ImmutableMap.Builder<K, Set<Provider<V>>> providerMultimapBuilder =
+              ImmutableMap.builder();
+          for (Entry<K, ImmutableSet.Builder<Provider<V>>> entry
+              : providerMultimapMutable.entrySet()) {
+            providerMultimapBuilder.put(entry.getKey(), entry.getValue().build());
+          }
+          providerMultimap = providerMultimapBuilder.build();
+        }
+
+        @Override public Map<K, Set<Provider<V>>> get() {
+          return providerMultimap;
+        }
+
+        @Override public Set<Dependency<?>> getDependencies() {
+          return dependencies;
+        }
+      }
+
+      final class RealMultimapProvider
+          extends RealMapBinderProviderWithDependencies<Map<K, Set<V>>> {
+        private final ImmutableSet<Dependency<?>> dependencies;
+        private final Provider<Map<K, Set<Provider<V>>>> multimapProvider;
+
+        RealMultimapProvider(
+            ImmutableSet<Dependency<?>> dependencies,
+            Provider<Map<K, Set<Provider<V>>>> multimapProvider) {
+          super(multimapKey);
+          this.dependencies = dependencies;
+          this.multimapProvider = multimapProvider;
+        }
+
+        @Override public Map<K, Set<V>> get() {
+          ImmutableMap.Builder<K, Set<V>> multimapBuilder = ImmutableMap.builder();
+          for (Entry<K, Set<Provider<V>>> entry : multimapProvider.get().entrySet()) {
+            K key = entry.getKey();
+            ImmutableSet.Builder<V> valuesBuilder = ImmutableSet.builder();
+            for (Provider<V> valueProvider : entry.getValue()) {
+              V value = valueProvider.get();
+              checkConfiguration(value != null,
+                  "Multimap injection failed due to null value for key \"%s\"", key);
+              valuesBuilder.add(value);
+            }
+            multimapBuilder.put(key, valuesBuilder.build());
+          }
+          return multimapBuilder.build();
+        }
+
+        @Override public Set<Dependency<?>> getDependencies() {
+          return dependencies;
+        }
+      }
     }
 
     /**
-     * Implementation of {@code {@link Map.Entry}&lt;K, {@link Provider}&lt;V&gt;&gt;} that
-     * defines equality based on the provider's key rather than the provider itself. This allows
-     * Guice to remove duplicate bindings.
-     *
-     * @param <K> the map's key type
-     * @param <V> the type provided by the map's values
+     * A Provider that Map.Entry that is also a Provider.  The key is the entry in the
+     * map this corresponds to and the value is the provider of the user's binding.
+     * This returns itself as the Provider.get value.
      */
-    private static final class ProviderMapEntry<K, V> implements
+    static final class ProviderMapEntry<K, V> implements
         ProviderWithDependencies<Map.Entry<K, Provider<V>>>, Map.Entry<K, Provider<V>> {
       private final K key;
       private final Provider<V> provider;
-      private volatile Key<V> valueKey;
+      private final Key<V> valueKey;
 
       private ProviderMapEntry(K key, Provider<V> provider, Key<V> valueKey) {
         this.key = key;
         this.provider = provider;
         this.valueKey = valueKey;
       }
-      
-      public Entry<K, Provider<V>> get() {
+
+      @Override public Entry<K, Provider<V>> get() {
         return this;
       }
-      
-      public Set<Dependency<?>> getDependencies() {
+
+      @Override public Set<Dependency<?>> getDependencies() {
         return ((HasDependencies) provider).getDependencies();
       }
 
       public Key<V> getValueKey() {
-        // Every time, check if the value key needs rehashing.
-        // If so, update the field as an optimization for next time.
-        Key<V> currentValueKey = valueKey;
-        if (needsRehashing(currentValueKey)) {
-          currentValueKey = rehash(currentValueKey);
-          valueKey = currentValueKey;
-        }
-        return currentValueKey;
+        return valueKey;
       }
 
-      public K getKey() {
+      @Override public K getKey() {
         return key;
       }
 
-      public Provider<V> getValue() {
+      @Override public Provider<V> getValue() {
         return provider;
       }
 
-      public Provider<V> setValue(Provider<V> value) {
+      @Override public Provider<V> setValue(Provider<V> value) {
         throw new UnsupportedOperationException();
       }
 
-      /**
-       * We don't follow the normal rules for Map.Entry equality, but that's okay, because
-       * the providers we return will never compare equal to anything else anyway. By defining
-       * equality this way, we let Guice remove duplicate bindings.
-       */
-      @Override public boolean equals(Object obj) {
-        return obj instanceof ProviderMapEntry
-            && key.equals(((ProviderMapEntry<?, ?>) obj).getKey())
-            && getValueKey().equals(((ProviderMapEntry<?, ?>) obj).getValueKey());
+       @Override public boolean equals(Object obj) {
+         if (obj instanceof Map.Entry) {
+           Map.Entry o = (Map.Entry)obj;
+           return Objects.equal(key, o.getKey())
+               && Objects.equal(provider, o.getValue());
+         }
+         return false;
       }
 
-      /**
-       * Again, we are not following the normal rules for Map.Entry, this time to avoid
-       * our hashCode changing if our {@link #valueKey} does.
-       */
       @Override public int hashCode() {
-        return key.hashCode();
+        return key.hashCode() ^ provider.hashCode();
       }
 
       @Override public String toString() {
-        return "ProviderMapEntry(" + key + ", " + valueKey + ")";
+        return "ProviderMapEntry(" + key + ", " + provider + ")";
       }
     }
 
